/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    armarx_control::ArmarXObjects::NJointControllerExample
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once


#include <VirtualRobot/Robot.h>

#include <ArmarXCore/util/tasks.h>

#include <RobotAPI/libraries/armem/client/plugins.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include <armarx/control/client/ComponentPlugin.h>
#include <armarx/control/client/ControllerBuilder.h>
#include <armarx/control/common/type.h>
#include <armarx/control/njoint_controller/controller_descriptions.h>


namespace armarx::control::components::example_client
{

    inline const control::common::ControllerType ControllerType =
        control::common::ControllerType::TSImpedance;

    /**
     * @defgroup Component-ExampleClient ExampleClient
     * @ingroup armarx_navigation-Components
     * A description of the component ExampleClient.
     *
     * @class ExampleClient
     * @ingroup Component-ExampleClient
     * @brief Brief description of class ExampleClient.
     *
     * Detailed description of class ExampleClient.
     */
    class Component :
        virtual public armarx::Component,
        virtual public armarx::control::client::ComponentPluginUser,
        virtual public armem::ListeningClientPluginUser
    {

    public:
        Component();

        ~Component() override;

    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;

        std::string getDefaultName() const override;


        void runExample();

    private:
        void goalReached();

        struct
        {
            std::string robotName = "Armar6";
            std::string rns = "RightArm";

            std::string configName = "default";

            float movementRadius = 200; // [mm]
            float movementDurationSeconds = 20; // [s]
        } properties;

        armarx::RunningTask<Component>::pointer_type task;

        armem::client::plugins::ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>*
            virtualRobotReaderPlugin = nullptr;

        VirtualRobot::RobotPtr robot;

        std::optional<control::client::ControllerWrapper<ControllerType>> ctrl;
    };

} // namespace armarx::control::components::example_client
