#include "aron_conversions.h"
#include <armarx/control/njoint_mp_controller/task_space/ImpedanceController.h>
#include <armarx/control/njoint_mp_controller/task_space/AdmittanceController.h>
#include <armarx/control/njoint_mp_controller/task_space/KVILImpedanceController.h>


namespace armarx::control::njoint_mp_controller::task_space
{
    void
    fromAron(const arondto::TSImpedanceMPConfig& dto,
             TSImpedanceMPConfig& bo)
    {
        fromAron(static_cast<const armarx::control::common::control_law::arondto::TaskspaceImpedanceControllerConfig &>(dto),
                 static_cast<armarx::control::common::control_law::TaskspaceImpedanceController::Config &>(bo));
        fromAron(dto.mpConfig, bo.mpConfig);
    }

    void
    fromAron(const arondto::TSAdmittanceMPConfig& dto,
             TSAdmittanceMPConfig& bo)
    {
        fromAron(static_cast<const armarx::control::common::control_law::arondto::TaskspaceAdmittanceControllerConfig &>(dto),
                 static_cast<armarx::control::common::control_law::TaskspaceAdmittanceController::Config &>(bo));
        fromAron(dto.mpConfig, bo.mpConfig);
    }

    void
    fromAron(const arondto::Constraint& dto,
             Constraint& bo)
    {
        bo = Constraint{
            .type = dto.type,
            .pcaMean = dto.pcaMean,
            .pcaComponents = dto.pcaComponents,
            .localFrameRotMat = dto.localFrameRotMat,
            .localFrameOrigin = dto.localFrameOrigin,
            .vmpStart = dto.vmpStart,
            .priority = dto.priority,
            .density_mu = dto.density_mu,
            .density_weights = dto.density_weights,
            .density_std_dev = dto.density_std_dev,
            .spatial_scale = dto.spatial_scale,
            .pm_scale = dto.pm_scale,
            .pm_init_proj_point = dto.pm_init_proj_point,
            .pm_init_proj_value = dto.pm_init_proj_value,
            .pm_init_tangent_vec = dto.pm_init_tangent_vec
        };
    }

    Constraint
    fromAron(const arondto::Constraint& dto)
    {
        Constraint bo = Constraint{
                .type = dto.type,
                .pcaMean = dto.pcaMean,
                .pcaComponents = dto.pcaComponents,
                .localFrameRotMat = dto.localFrameRotMat,
                .localFrameOrigin = dto.localFrameOrigin,
                .vmpStart = dto.vmpStart,
                .priority = dto.priority,
                .density_mu = dto.density_mu,
                .density_weights = dto.density_weights,
                .density_std_dev = dto.density_std_dev,
                .spatial_scale = dto.spatial_scale,
                .pm_scale = dto.pm_scale,
                .pm_init_proj_point = dto.pm_init_proj_point,
                .pm_init_proj_value = dto.pm_init_proj_value,
                .pm_init_tangent_vec = dto.pm_init_tangent_vec
        };
        return bo;
    }

    void
    fromAron(const arondto::KVILImpedanceMPConfig& dto,
             KVILImpedanceMPConfig& bo)
    {
        fromAron(static_cast<const armarx::control::common::control_law::arondto::KeypointControllerConfig &>(dto),
                 static_cast<armarx::control::common::control_law::KeypointsImpedanceController::Config &>(bo));
        fromAron(dto.mpConfig, bo.mpConfig);
        std::transform(dto.constraintList.begin(), dto.constraintList.end(), std::back_inserter(bo.constraintList),
                       [](const auto& c) -> Constraint
                       {
                           return fromAron(c);
                       });
    }

} /// namespace armarx::control::njoint_mp_controller::task_space
