#pragma once

#include <RobotAPI/libraries/skills/provider/SpecializedSkill.h>

#include <armarx/control/skills/skills/aron/ShapeHandParams.aron.generated.h>

#include "../skills.h"

namespace armarx::control::skills::skills
{

    class ShapeHand : public ::armarx::skills::SpecializedSkill<ShapeHandParams>
    {

    public:
        static ::armarx::skills::SkillDescription
        GetSkillDescription()
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = "ShapeHand"},
                .description =
                    "Shapes the hand according to a predefined shape or passed joint angles",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        ShapeHand(const HandUnitRemote&);

    private:
        ::armarx::skills::Skill::MainResult main() override;
        Skill::MainResult waitUntilFinished(const NameValueMap& targetAngles);

    private:
        HandUnitRemote remote;

        float getAccuracy(const std::string& jointName);
    };

    class SimpleShapeHand : public ::armarx::skills::SpecializedSkill<SimpleShapeHandParams>
    {
    public:
        static ::armarx::skills::SkillDescription
        GetSkillDescription(const std::string& shapeName)
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = shapeName + "Hand"},
                .description = std::string("Sets shape of hand to '") + shapeName + "'",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        SimpleShapeHand(const std::string& shapeName);

    private:
        ::armarx::skills::Skill::MainResult main() override;

    private:
        const std::string shapeName;
    };

} // namespace armarx::control::skills::skills
